#version 330

smooth in vec2 vTexCoord;
smooth in vec3 vNormal;
smooth in vec3 vWorldPos;
smooth in vec3 vTangent;
smooth in vec4 shadowCoordinates;
smooth in vec4 projectedTextureCoordinates;

out vec4 outputColor;

uniform sampler2D gSampler;
uniform sampler2D gNormalMap; 
uniform sampler2D shadowMap; 
uniform sampler2D projectedTexture; 

uniform vec4 vColor;
uniform vec3 vCameraPos;
uniform int normalMappingOn; // Is normal mapping active?
uniform int shadowsOn; // Is shadow mapping active?
uniform int projectedTextureOn; // Is projectedTexture active?

struct SimpleDirectionalLight{
	vec3 vColor;
	vec3 vDirection;
	float fAmbientIntensity;
};
uniform SimpleDirectionalLight sunLight;

struct Material
{
   float specularIntensity;
   float specularPower;
};
uniform Material material;

// Method declarations
vec4 getDirectionalLightColor(SimpleDirectionalLight dirLight, vec3 vNormal);
vec4 getSpecularColor(vec3 vWorldPos, vec3 vCameraPos, vec3 vNormal, Material mat, SimpleDirectionalLight dLight);
vec3 calcBumpedNormal();
float getVisibility(sampler2D shadowMap, vec4 vShadowCoord);

// ********************************************************************
// Method definitions
void main()
{
	vec4 vTexColor = texture2D(gSampler, vTexCoord);
	vec4 vMixedColor = vTexColor * vColor;

	vec4 projColour = vec4(1.0);
	if (projectedTextureOn == 1) {
		projColour = texture2D(projectedTexture, projectedTextureCoordinates.xy) * vColor;
	}

	vec3 vNNormal  = normalize(vNormal);
	if(normalMappingOn == 1) {	
		vNNormal = calcBumpedNormal();
	} else {
		vNNormal  = normalize(vNormal);
	}
	
	vec4 vDiffuseColor = getDirectionalLightColor(sunLight, vNNormal);
	vec4 vSpecularColor = getSpecularColor(vWorldPos, vCameraPos, vNNormal, material, sunLight);
	
	float visibility = getVisibility(shadowMap, shadowCoordinates);

	outputColor = (vMixedColor * vDiffuseColor + vSpecularColor) * visibility;
	
	if (projectedTextureOn == 1) {
		if ((clamp(projectedTextureCoordinates.x, 0.0, 1.0) == projectedTextureCoordinates.x) && 
			(clamp(projectedTextureCoordinates.y, 0.0, 1.0) == projectedTextureCoordinates.y)) {
			outputColor = outputColor + (projColour * vDiffuseColor * visibility);
		}
	}
}

vec4 getDirectionalLightColor(SimpleDirectionalLight dirLight, vec3 vNormal)
{
	float fDiffuseIntensity = max(0.0f, dot(vNormal, -dirLight.vDirection));
	float fMult = clamp(dirLight.fAmbientIntensity + fDiffuseIntensity, 0.0f, 1.0f);
	return vec4(dirLight.vColor * fMult, 1.0f);
}

vec4 getSpecularColor(vec3 vWorldPos, vec3 vCameraPos, vec3 vNormal, Material mat, SimpleDirectionalLight dLight)
{
   vec4 vResult = vec4(0.0, 0.0, 0.0, 0.0);
   
   vec3 vReflectedVector = normalize(reflect(dLight.vDirection, vNormal));
   vec3 vVertexToEyeVector = normalize(vCameraPos - vWorldPos);
   float fSpecularFactor = dot(vVertexToEyeVector, vReflectedVector);
   
   if (fSpecularFactor > 0)
   {
      fSpecularFactor = pow(fSpecularFactor, mat.specularPower);
      vResult = vec4(dLight.vColor, 1.0) * mat.specularIntensity * fSpecularFactor;  
   }
   return vResult;
}

vec3 calcBumpedNormal()
{
    vec3 Normal = normalize(vNormal);
    vec3 Tangent = normalize(vTangent);
    Tangent = normalize(Tangent - dot(Tangent, Normal) * Normal);
    vec3 Bitangent = cross(Tangent, Normal);
    vec3 BumpMapNormal = texture(gNormalMap, vTexCoord).xyz;
    BumpMapNormal = 2.0 * BumpMapNormal - vec3(1.0, 1.0, 1.0);
    mat3 TBN = mat3(Tangent, Bitangent, Normal);
    vec3 newNormal = TBN * BumpMapNormal;
    newNormal = normalize(newNormal);
    return newNormal;
}


// Shadow mapping:

vec2 poissonDisk[16] = vec2[]( 
   vec2( -0.94201624, -0.39906216 ), 
   vec2( 0.94558609, -0.76890725 ), 
   vec2( -0.094184101, -0.92938870 ), 
   vec2( 0.34495938, 0.29387760 ), 
   vec2( -0.91588581, 0.45771432 ), 
   vec2( -0.81544232, -0.87912464 ), 
   vec2( -0.38277543, 0.27676845 ), 
   vec2( 0.97484398, 0.75648379 ), 
   vec2( 0.44323325, -0.97511554 ), 
   vec2( 0.53742981, -0.47373420 ), 
   vec2( -0.26496911, -0.41893023 ), 
   vec2( 0.79197514, 0.19090188 ), 
   vec2( -0.24188840, 0.99706507 ), 
   vec2( -0.81409955, 0.91437590 ), 
   vec2( 0.19984126, 0.78641367 ), 
   vec2( 0.14383161, -0.14100790 ) 
);

float getVisibility(sampler2D shadowMap, vec4 vShadowCoord)
{
    if(shadowsOn == 0) {
		return 1.0;
	
	}
    float visibility = 1.0;
    float bias = 0.005;
 
    vec4 shadowCoordinateWdivide = vec4(vShadowCoord.x/vShadowCoord.w, vShadowCoord.y/vShadowCoord.w, (vShadowCoord.z-bias)/vShadowCoord.w, 1.0) ;

    for (int i=0;i<4;i++)
    {
      int index = i;
		  vec4 vShadowSmooth = vec4(vShadowCoord.x + poissonDisk[index].x/700.0, vShadowCoord.y + poissonDisk[index].y/700.0, (vShadowCoord.z-bias)/vShadowCoord.w, 1.0);
		  float fSub = texture(shadowMap, vShadowSmooth.st).r; 
		  visibility -= 0.2*(1.0-fSub);
    }

    return pow(visibility, 6) * 10;
}